"""
A module for every constant of symmetric.
"""


# Logs
LOG_FILE_NAME = "symmetric.log"

# API token authentication
API_CLIENT_TOKEN_NAME = "symmetric_api_key"
API_SERVER_TOKEN_NAME = "SYMMETRIC_API_KEY"
API_DEFAULT_TOKEN = "symmetric_token"

# Route parsing
ROUTE_PATTERN = (r"^\/$|^(?!.*[\/\-]{2,}.*)(?!.*[\-_]{2,}.*)(?!.*_{2,}.*)\/"
                 r"[a-zA-Z\_\-\/]*(?<!\-)(?<!_)(?<!\/)$")
# ^\/$  => Match a "/" string. (the "|" operator means or)
# ^(?!.*[\/\-]{2,}.*)  => Don't allow repetition of "/" or "-" characters or
#                         combinations of both, like "/-"
# (?!.*[\-_]{2,}.*)  => Don't allow repetition of "-" or "_" characters or
#                       combinations of both, like "-_"
# (?!.*_{2,}.*)  => Don't allow repetition of "_"
# \/  => The route must always start with a "/"
# [a-zA-Z\_\-\/]*  => Then, any amount of letters, hyphens, underscores and/or
#                     slashes can be allowed (as long as the don't conflict)
#                     with previous rules, namely, the lookaheads)
# (?<!\-)(?<!_)(?<!\/)$  => The final character can't be a hyphen, an
#                           underscore nor a slash



"""
A module to set up the logging configuration of symmetric.
"""

import os
import logging.config

import symmetric.constants


# Logging configuration
logging.config.dictConfig({
    "version": 1,
    "formatters": {
        "console": {
            "format": "[%(asctime)s] [%(levelname)s] %(module)s: %(message)s"
        },
        "file": {
            "format": ("[%(asctime)s] [%(levelname)s] %(pathname)s - "
                       "line %(lineno)d: \n%(message)s\n")
        }
    },
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stderr",
            "formatter": "console"
        },
        "file": {
            "class": "logging.FileHandler",
            "filename": os.getenv(
                "LOG_FILE",
                default=symmetric.constants.LOG_FILE_NAME
            ),
            "formatter": "file"
        }
    },
    "root": {
        "level": "INFO",
        "handlers": ["console", "file"]
    }
})



"""
The main module of symmetric.
"""

import sys
import json
import bisect
import flask

import symmetric.logging
import symmetric.constants
import symmetric.endpoints
import symmetric.helpers
import symmetric.errors


class Symmetric:

    """
    Main class to encapsulate every important feature of the symmetric package.
    """

    # Define allowed HTTP methods
    __allowed_methods = [
        "OPTIONS",
        "GET",
        "HEAD",
        "POST",
        "PUT",
        "DELETE",
        "TRACE"
    ]

    def __init__(self, app_object):
        self.__app = app_object
        self.__endpoints = []
        self.__server_token_name = symmetric.constants.API_SERVER_TOKEN_NAME
        self.__client_token_name = symmetric.constants.API_CLIENT_TOKEN_NAME

    def __call__(self, *args, **kwargs):
        """
        Enable WSGI servers to start with CLI utilities
        (like gunicorn module:symmetric).
        """
        return self.__app.__call__(*args, **kwargs)

    def router(self, route, methods=["get"], response_code=200,
               auth_token=False):
        """
        Decorator modifier. Recieves a route string, a list of HTTP methods, a
        response code and a boolean indicating whether or not to authenticate.
        If the route does not start with '/', it gets added. HTTP methods are
        also filtered. Returns the final decorated function.
        """
        try:
            symmetric.helpers.parse_route(route)
        except symmetric.errors.IncorrectRouteFormatError as err:
            self.__app.logger.error(
                f"[[symmetric]] IncorrectRouteFormatError: {err}"
            )
            sys.exit(1)

        methods = [
            symmetric.helpers.verb(x) for x in methods
            if symmetric.helpers.verb(x) in Symmetric.__allowed_methods
        ]

        def decorator(function):
            """
            Function decorator. Recieves the main function and wraps it as a
            flask endpoint. Returns the wrapped function.
            """
            try:
                self.__save_endpoint(
                    symmetric.endpoints.Endpoint(
                        route, methods, response_code, function, auth_token
                    )
                )
            except symmetric.errors.DuplicatedRouteError as err:
                self.__app.logger.error(
                    f"[[symmetric]] DuplicatedRouteError: {err}"
                )
                sys.exit(1)

            # Decorate the wrapper
            @self.__app.route(
                route, methods=methods, endpoint=function.__name__
            )
            def wrapper(*args, **kwargs):
                """
                Function wrapper. The main function gets logged, the JSON body
                gets extracted from the request and gets unpacked as **kwargs
                to pass to the main function. Some precautions are also taken
                (namely a try/except combo). Returns the function's output
                jsonified with a response code.
                """
                try:
                    self.__log_request(flask.request, route, function)

                    # Get the body
                    body = flask.request.get_json()
                    if not body:
                        body = {}

                    # Check for token authentication
                    symmetric.helpers.authenticate(
                        body, auth_token, self.__client_token_name,
                        self.__server_token_name)

                    # Filter method parameters
                    parameters = symmetric.helpers.filter_params(
                        function, body, auth_token, self.__client_token_name)
                    return flask.jsonify(function(**parameters)), response_code
                except symmetric.errors.AuthenticationRequiredError as err:
                    # Error authenticating
                    self.__app.logger.error(
                        f"[[symmetric]] exception caught: {err}"
                    )
                    return flask.jsonify({}), 401
                except Exception as err:
                    self.__app.logger.error(
                        f"[[symmetric]] exception caught: {err}"
                    )
                    return flask.jsonify({}), 500
            return wrapper
        return decorator

    def run(self, *args, **kwargs):
        """Executes the main run function of the Flask object."""
        self.__app.run(*args, **kwargs)

    def generate_documentation(self, module_name):
        """
        Gets the documentation of every endpoint and assembles it into a
        markdown formatted string.
        """
        docs = f"# {symmetric.helpers.humanize(module_name)} "
        docs += "API Documentation\n\n"
        docs += ("Endpoints that require an authentication token should "
                 f"send it in a key named `{self.__client_token_name}` "
                 "inside the request body.\n\n")
        raw_docs = [x.generate_documentation() for x in self.__endpoints]
        docs += "\n".join(raw_docs)
        return docs

    def __save_endpoint(self, endpoint):
        """Saves an endpoint object and sorts the endpoints list."""
        if endpoint in self.__endpoints:
            message = f"Endpoint '{endpoint.route}' was defined twice."
            raise symmetric.errors.DuplicatedRouteError(message)
        bisect.insort(self.__endpoints, endpoint)

    def __log_request(self, request, route, function):
        """
        Recieves a request object, a route string and a function and
        logs the request event. It also logs the json body of the request.
        """
        self.__app.logger.info(
            f"{request.method} request to '{route}' endpoint "
            f"('{function.__name__}' function)."
        )
        body = request.get_json()
        if body:
            self.__log_body(body)

    def __log_body(self, body):
        """Given a json request body, logs it."""
        self.__app.logger.info("Request Body:\n" + json.dumps(
            body,
            indent=2,
            sort_keys=False,
            ensure_ascii=False
        ))


# Create flask app object
app = flask.Flask(__name__)


# Create symmetric object
sym = Symmetric(app)



"""
A module to hold some CLI utilities.
"""

import os
import sys
import traceback
import importlib

import symmetric.errors


def start_server(module, server, port, debug):
    """
    Gets the symmetric object and then runs it with the parameters given
    to the method.
    """
    symmetric_object = get_symmetric_object(module, debug)
    symmetric_object.run(host=server, port=port, debug=debug)


def document_api(module, filename):
    """
    Gets the symmetric object and then calls the documentation method.
    """
    symmetric_object = get_symmetric_object(module, True)
    docs = symmetric_object.generate_documentation(module)
    with open(filename, "w") as docs_file:
        docs_file.write(docs)


def get_symmetric_object(module_name, debug):
    """
    Imports the module :module_name and the tries to find the
    symmetric object. This method is strongly inspired in gunicorn's own
    import_app method and uvicorn's main method.
    """
    # Import the module
    try:
        # Add current directory to path
        sys.path.insert(0, ".")
        module = importlib.import_module(module_name)
    except ImportError:
        # If the user wrote module.py instead of just module
        if module_name.endswith(".py") and os.path.exists(module_name):
            actual_name = module_name.rsplit('.', 1)[0]
            error = (f"Module {module_name} not found. "
                     f"Did you mean {actual_name}?")
            raise ImportError(error)
        raise

    # Get the symmetric object
    try:
        symmetric_object = getattr(module, "symmetric")
    except AttributeError:
        if debug:
            traceback.print_exception(*sys.exc_info())
        error = f"Failed to find the symmetric object in {module_name}."
        raise symmetric.errors.AppImportError(error)

    return symmetric_object



"""
A module for every helper of symmetric.
"""

import re
import os
import inspect

import symmetric.constants
import symmetric.errors


def verb(dirty):
    """
    Given a 'dirty' string (lowercased, with trailing whitespaces), strips
    it and returns it uppercased.
    """
    return dirty.strip().upper()


def humanize(module_name):
    """Transforms a module name into a pretty human-likable string."""
    module_name = module_name.lower()
    module_name = module_name.replace('_', ' ').replace('-', ' ')
    module_name = module_name.title()
    return module_name


def parse_route(route):
    """
    If :route does not match the expected route pattern,
    raises IncorrectRouteFormatError.
    """
    if re.fullmatch(symmetric.constants.ROUTE_PATTERN, route) is None:
        message = (f"Your route '{route}' does not match with the symmetric "
                   "route guidelines. Refer to the documentation at "
                   "https://github.com/daleal/symmetric for more information.")
        raise symmetric.errors.IncorrectRouteFormatError(message)


def authenticate(body, auth_token, client_token_name, server_token_name):
    """
    Raises an exception if the body does not include the client token
    or if it is different to the server token.
    """
    if not auth_token:
        # No auth is required
        return
    # Auth is required from now on
    if client_token_name not in body:
        # The body does not include the desired token
        error = "The request does not include an authentication token."
        raise symmetric.errors.AuthenticationRequiredError(error)
    # If the token in the body equals the one in the env, return True
    token = os.getenv(server_token_name, symmetric.constants.API_DEFAULT_TOKEN)
    if body[client_token_name] != token:
        error = "Incorrect authentication token."
        raise symmetric.errors.AuthenticationRequiredError(error)


def filter_params(function, data, has_token, token_key):
    """Filters parameters so that the function recieves only what it needs."""
    # Filter token key
    if has_token:
        data.pop(token_key, None)

    # Get the parameters
    params = inspect.getfullargspec(function)
    if params.varkw is not None:
        # The function recieves kwargs, return the full dictionary
        return data
    if not params.args:
        # The function does not recieve args, return an empty dict
        return {}
    # Filter every param whose key is not in the params dictionary
    return {k: v for k, v in data.items() if k in params.args}



"""
A module to route the CLI traffic.
"""

import sys
import argparse

import symmetric.cli_utils


def dispatcher():
    """
    Main CLI method, recieves the command line action and dispatches it to
    the corresponding method.
    """
    parser = generate_parser()

    args = parser.parse_args()

    try:
        if args.action == "run":
            symmetric.cli_utils.start_server(
                args.module, args.server, args.port, args.debug
            )
        elif args.action == "docs":
            symmetric.cli_utils.document_api(args.module, args.filename)
    except AttributeError:
        print("An argument is required for the symmetric command.")
        parser.print_help()
        sys.exit(1)


def generate_parser():
    """Generates the CLI parser for the module."""
    # Create parser
    parser = argparse.ArgumentParser(
        description="Command line interface tool for symmetric."
    )

    # In order to allow the CLI utility grow, the parser will include an
    # initial argumment to determine which subparser will be executed.

    # Create subparsers
    subparsers = parser.add_subparsers(help="Action to be executed.")

    # Runner parser
    generate_runner_subparser(subparsers)

    # Documentation parser
    generate_documentation_subparser(subparsers)

    return parser


def generate_runner_subparser(subparsers):
    """Generates the subparser for the run server option."""
    runner_parser = subparsers.add_parser("run")
    runner_parser.set_defaults(action="run")

    # Module name
    runner_parser.add_argument(
        "module",
        metavar="module",
        help="Name of the module that uses the symmetric object."
    )

    # Host
    runner_parser.add_argument(
        "-s", "--server",
        dest="server",
        default="127.0.0.1",
        help="Server hostname in which the application will run."
    )

    # Port
    runner_parser.add_argument(
        "-p", "--port",
        dest="port",
        type=int,
        default=5000,
        help="Port for the webserver."
    )

    # Debug
    runner_parser.add_argument(
        "-d", "--no-debug",
        dest="debug",
        action='store_const',
        default=True,  # debug is set to True by default
        const=False,   # if the flag is used, sets debug to False
        help="Do not run in debug mode."
    )


def generate_documentation_subparser(subparsers):
    """Generates the subparser for the auto-documentation option."""
    documentation_parser = subparsers.add_parser("docs")
    documentation_parser.set_defaults(action="docs")

    # Module name
    documentation_parser.add_argument(
        "module",
        metavar="module",
        help="Name of the module that uses the symmetric object."
    )

    # Filename
    documentation_parser.add_argument(
        "-f", "--filename",
        dest="filename",
        default="documentation.md",
        help="Name of the file in where to write the documentation."
    )


if __name__ == "__main__":
    dispatcher()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['symmetric']

package_data = \
{'': ['*']}

install_requires = \
['flask>=1.1.1,<2.0.0']

entry_points = \
{'console_scripts': ['symmetric = symmetric.command_line_interface:dispatcher']}

setup_kwargs = {
    'name': 'symmetric',
    'version': '2.1.1',
    'description': 'A powerful yet lean wrapper over Flask to massively speed up API creations and enable super fast module-to-API transformations.',
    'long_description': '# Symmetric\n\n![PyPI - Downloads](https://img.shields.io/pypi/dm/symmetric?style=for-the-badge&logo=python&color=306998)\n\nA powerful yet lean wrapper over **[Flask](https://github.com/pallets/flask)** to massively speed up **[API](https://en.wikipedia.org/wiki/Web_API)** creations and enable super fast module-to-**[API](https://en.wikipedia.org/wiki/Web_API)** transformations.\n\n![GitHub Workflow Status](https://img.shields.io/github/workflow/status/daleal/symmetric/tests?label=tests&logo=github&style=for-the-badge)\n![GitHub Workflow Status](https://img.shields.io/github/workflow/status/daleal/symmetric/linters?label=linters&logo=github&style=for-the-badge)\n\n## Why Symmetric?\n\nRaw developing speed and ease of use, that\'s why. While `Flask` is a powerful tool to have, getting it to work from scratch can be a bit of a pain, especially if you have never used it before. The idea behind `symmetric` is to be able to take any module **already written** and transform it into a working API in a matter of minutes, instead of having to design the module ground-up to work with `Flask` (it can also be used to build an API from scratch really fast). With `symmetric`, you will also get some neat features, namely:\n\n- Auto logging.\n- Server-side error detection and exception handling.\n- Native support for an authentication token on a per-endpoint basis.\n- Auto-generated documentation for your API.\n\n## Installing\n\nInstall using pip!\n\n```bash\npip install symmetric\n```\n\n## Usage\n\n### Running the development server\n\nTo start the development server, just run:\n\n```bash\nsymmetric run <module>\n```\n\nWhere `<module>` is your module name (in the examples, we will be writing in a file named `module.py`, so the module name will be just `module`). A `Flask` instance will be spawned immediately and can be reached at [http://127.0.0.1:5000](http://127.0.0.1:5000) by default. We don\'t have any endpoints yet, so we\'ll add some later. **Do not use this in production**. The `Flask` server is meant for development only. Instead, you can use any `WSGI` server to run the API. For example, to run the API using [gunicorn](https://gunicorn.org/), you just need to run `gunicorn module:symmetric` and a production ready server will be spawned.\n\n### Defining the API endpoints\n\nThe module consists of a main object called `symmetric`, which includes an important element: the `router` decorator. Let\'s analyze it:\n\n```py\nfrom symmetric import symmetric\n\n@symmetric.router("/some-route", methods=["get"], response_code=200, auth_token=False)\n```\n\nThe decorator recieves 4 arguments: the `route` argument (the endpoint of the API to which the decorated function will map), the `methods` argument (a list of the methods accepted to connect to that endpoint, defaults in only `GET` requests), the `response_code` argument (the response code of the endpoint if everything goes according to the plan. Defaults to `200`) and the `auth_token` argument (a boolean stating if the endpoint requires authentication using a `symmetric` token. Defaults to `False`).\n\nNow let\'s imagine that we have the following method:\n\n```py\ndef some_function():\n    """Greets the world."""\n    return "Hello World!"\n```\n\nTo transform that method into an API endpoint, all you need to do is add one line:\n\n```py\n@symmetric.router("/sample")\ndef some_function():\n    """Greets the world."""\n    return "Hello World!"\n```\n\nRun `symmetric run module` and send a `GET` request to `http://127.0.0.1:5000/sample`. You should get a `Hello World!` in response! (To try it with a browser, make sure to run the above command and click [this link](http://127.0.0.1:5000/sample)).\n\nBut what about methods with arguments? Of course they can be API\'d too! Let\'s now say that you have the following function:\n\n```py\ndef another_function(a, b=372):\n    """\n    Adds :a and :b and returns the result of\n    that operation.\n    """\n    return a + b\n```\n\nTo transform that method into an API endpoint, all you need to do, again, is add one line:\n\n```py\n@symmetric.router("/add")\ndef another_function(a, b=372):\n    """\n    Adds :a and :b and returns the result of\n    that operation.\n    """\n    return a + b\n```\n\n### Querying API endpoints\n\nTo give parameters to a function, all we need to do is send a `json` body with the names of the parameters as keys. Let\'s see how! Run `symmetric run module` and send a `GET` request to `http://127.0.0.1:5000/add`, now using the `requests` module.\n\n```python\nimport requests\n\npayload = {\n    "a": 48,\n    "b": 21\n}\nresponse = requests.get("http://127.0.0.1:5000/add", json=payload)\nprint(response.json())\n```\n\nWe got a `69` response! (`48 + 21 = 69`). Of course, you can return dictionaries from your methods and those will get returned as a `json` body in the response object **automagically**!\n\nWith this in mind, you can transform any existing project into a usable API very quickly!\n\n### The `symmetric` token authentication\n\nTo speed up your API creation even more, `symmetric` includes native support for a simple token authentication. It works like this:\n\n1. **Set up the token in the server.**\n\n    In the environment where your API is going to run, add an environmental variable named `SYMMETRIC_API_KEY` and set its value to be the _[pre-shared](https://en.wikipedia.org/wiki/Pre-shared_key)_ token. If you don\'t set the environmental key, the _default_ `SYMMETRIC_API_KEY` value will be `symmetric_token` (in your development environment that\'s probably fine, but in the production server you should **never** use the default value of the `symmetric` token).\n\n2. **Force one of your endpoints to use an authentication token.**\n\n    Let\'s say your module has a method like this:\n\n    ```py\n    def secret_function():\n        """Greets the world (secretly)."""\n        return "Hello World in secret!"\n    ```\n\n    Add the `symmetric` router decorator in the following manner:\n\n    ```py\n    @symmetric.router("/secret", auth_token=True)\n    def secret_function():\n        """Greets the world (secretly)."""\n        return "Hello World in secret!"\n    ```\n\n    Now, your endpoint won\'t respond to any request that is not correctly authenticated.\n\n3. **Query your endpoint.**\n\n    To query your endpoint, the request body must include a key named `symmetric_api_key` with a value to match the one of the environment\'s `SYMMETRIC_API_KEY`. So, for instance, if you are using the default `SYMMETRIC_API_KEY` value (`symmetric_token`), the request body for the `/secrets` endpoint should be:\n\n    ```py\n    payload = {\n        "symmetric_api_key": "symmetric_token"\n    }\n    ```\n\n    By sending that payload in the request body as a `json`, the request can access the endpoint correctly.\n\n### Auto-generating the API documentation\n\nGenerating API documentation is simple with `symmetric`. Just run the following command:\n\n```bash\nsymmetric docs <module>\n```\n\nThis will **automagically** generate a markdown file documenting each endpoint with the function docstring, required arguments and `HTTP` methods. Seems too simple to be true, right? Go ahead, try it yourself!\n\nYou can also specify the name of the documentation file (defaults to `documentation.md`) using the `-f` or the `--filename` flag.\n\n### The whole example\n\nTo sum up, if the original `module.py` file looked like this before `symmetric`:\n\n```py\ndef some_function():\n    """Greets the world."""\n    return "Hello World!"\n\n\ndef another_function(a, b=372):\n    """\n    Adds :a and :b and returns the result of\n    that operation.\n    """\n    return a + b\n\n\ndef secret_function():\n    """Greets the world (secretly)."""\n    return "Hello World in secret!"\n```\n\nThe complete final `module.py` file with `symmetric` should look like this:\n\n```py\nfrom symmetric import symmetric\n\n\n@symmetric.router("/sample")\ndef some_function():\n    """Greets the world."""\n    return "Hello World!"\n\n\n@symmetric.router("/add")\ndef another_function(a, b=372):\n    """\n    Adds :a and :b and returns the result of\n    that operation.\n    """\n    return a + b\n\n\n@symmetric.router("/secret", auth_token=True)\ndef secret_function():\n    """Greets the world (secretly)."""\n    return "Hello World in secret!"\n```\n\nTo run the server, just run `symmetric run module`. Now, you can send `GET` requests to `http://127.0.0.1:5000/sample`, `http://127.0.0.1:5000/add` and `http://127.0.0.1:5000/secret`. Here is a simple file to get you started querying your API:\n\n```py\nimport requests\n\n\ndef call_sample():\n    response = requests.get("http://127.0.0.1:5000/sample")\n    return response.text\n\n\ndef call_add():\n    payload = {\n        "a": 48,\n        "b": 21\n    }\n    response = requests.get("http://127.0.0.1:5000/add", json=payload)\n    return response.json()\n\n\ndef call_secret():\n    payload = {\n        "symmetric_api_key": "symmetric_token"\n    }\n    response = requests.get("http://127.0.0.1:5000/secret", json=payload)\n    return response.text\n\n\nif __name__ == \'__main__\':\n    print(call_sample())\n    print(call_add())\n    print(call_secret())\n```\n\nRunning `symmetric docs module` would result in a file `documentation.md` being created with the following content:\n\n``````pandoc\n# Module API Documentation\n\nEndpoints that require an authentication token should send it in a key named `symmetric_api_key` inside the request body.\n\n## `/add`\n\n### Description\n\nAdds :a and :b and returns the result of\nthat operation.\n\n### Metadata\n\n`HTTP` methods accepted: `GET`\n\nDoes not require an authentication token.\n\n### Parameters\n\n```py\n{\n    a,\n    b,  # defaults to 372\n}\n```\n\n## `/sample`\n\n### Description\n\nGreets the world.\n\n### Metadata\n\n`HTTP` methods accepted: `GET`\n\nDoes not require an authentication token.\n\n### Parameters\n\nNo required parameters.\n\n## `/secret`\n\n### Description\n\nGreets the world (secretly).\n\n### Metadata\n\n`HTTP` methods accepted: `GET`\n\nRequires an authentication token.\n\n### Parameters\n\nNo required parameters.\n\n``````\n\n## Logging\n\nBy default, the logs in the server will be written into the `stdout` and into a file named `symmetric.log`. You can change the name of the file by specifying the `LOG_FILE` environmental variable, if you want to.\n\n## Route rules\n\nThere are some rules regarding the correct routes that can be used. Failing to follow the `symmetric` route rules will result in the API not being run and an error being thrown and logged. To follow the rules, a route:\n\n1. **Can\'t** be defined twice.\n2. **Can\'t** have repetitions of `/`, `-` or `_`.\n3. **Can\'t** have concatenations of `/` with `-` or of `_` with `-`.\n4. **Can\'t** include characters other than letters (uppercase and/or lowercase), `/`, `-` and `_`.\n5. **Can\'t** end with a `/`, a `-` or a `_`. The **only** exception of this rule is when the route is just `/`, in which case, it **can** end with `/`.\n6. **Must** start with a `/`.\n\nHere are some examples.\n\n### Correct route patterns\n\n- `/`\n- `/symmetric`\n- `/hi/hello`\n- `/hello-world/basic_syntax`\n- `/_element/BIGelement`\n\n### Incorrect route patterns\n\n- `/hi//hello`\n- `element`\n- `/another-element/`\n- `/bad-_composition`\n- `/-worse`\n- `/element__two`\n- `/element2`\n- `/oof-number-one-`\n- `/oof_number_two_`\n\n## Developing\n\nClone the repository:\n\n```bash\ngit clone https://github.com/daleal/symmetric.git\n\ncd symmetric\n```\n\nRecreate environment:\n\n```bash\n./environment.sh\n\n. .venv/bin/activate\n```\n\nTest install:\n\n```bash\npoetry install  # will also install the symmetric CLI\n```\n\nRun the tests:\n\n```bash\npython -m unittest\n```\n\nBuild the project:\n\n```bash\npoetry build\n```\n\nPush to `TestPyPi`:\n\n```bash\npoetry config repositories.testpypi https://test.pypi.org/legacy/\npoetry publish -r testpypi\n```\n\nDownload from `TestPyPi`:\n\n```bash\ndeactivate\nrm -rf .testing-venv\npython3 -m venv .testing-venv\n. .testing-venv/bin/activate\npip install click\npython -m pip install --index-url https://test.pypi.org/simple/ symmetric\n```\n\nPush to `PyPi`:\n\n```bash\npoetry publish\n```\n',
    'author': 'Daniel Leal',
    'author_email': 'dlleal@uc.cl',
    'maintainer': 'Daniel Leal',
    'maintainer_email': 'dlleal@uc.cl',
    'url': 'https://github.com/daleal/symmetric',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



"""
This module contains the endpoints classes.
"""

import inspect


class Endpoint:

    """
    Class to encapsulate an endpoint.
    """

    def __init__(self, route, methods, response_code, function, has_token):
        self.__route = route
        self.__methods = methods
        self.__response_code = response_code
        self.__function = function
        self.__has_token = has_token

    def __lt__(self, other):
        return self.route < other.route

    def __eq__(self, other):
        return self.route == other.route

    @property
    def route(self):
        """Returns the route of the endpoint."""
        return self.__route

    def generate_documentation(self):
        """Generates the documentation of the function."""
        docstring = f"## `{self.__route}`\n\n"
        docstring += f"### Description\n\n"
        docstring += f"{self.__get_docstring()}\n\n"
        docstring += f"### Metadata\n\n"
        docstring += f"`HTTP` methods accepted: "
        docstring += f"{', '.join([f'`{x}`' for x in self.__methods])}\n\n"
        if self.__has_token:
            docstring += "Requires an authentication token.\n\n"
        else:
            docstring += "Does not require an authentication token.\n\n"
        docstring += f"### Parameters\n\n{self.__get_parameters()}\n"
        return docstring

    def __get_docstring(self):
        """Gets the docstring of the function."""
        docstring = inspect.getdoc(self.__function)
        return docstring if docstring else "No description provided."

    def __get_parameters(self):
        """Gets the parameters of the function."""
        params = inspect.getfullargspec(self.__function)
        if not params.args:
            return "No required parameters."
        parameters_amount = len(params.args)
        docstring = "```py\n{\n"
        if params.defaults:
            defaults_amount = len(params.defaults)
            no_defaults = parameters_amount - defaults_amount
            for ii in range(no_defaults):
                docstring += f"    {params.args[ii]},\n"

            for jj in range(defaults_amount):
                docstring += f"    {params.args[no_defaults + jj]},  # "
                if isinstance(params.defaults[jj], str):
                    default = f'"{params.defaults[jj]}"'
                else:
                    default = params.defaults[jj]
                docstring += f"defaults to {default}\n"
        else:
            for ii in range(parameters_amount):
                docstring += f"    {params.args[ii]},\n"

        docstring += "}\n```"

        return docstring



"""
Module to hold every custom exception.
"""


class AppImportError(Exception):
    """
    Exception for when there's an error finding the symmetric object inside
    the specified module.
    """


class AuthenticationRequiredError(Exception):
    """
    Exception for when an endpoint requires authentication and the request
    fails to provide the correct authentication token.
    """


class IncorrectRouteFormatError(Exception):
    """
    Exception for when an endpoint route is incorrectly defined (it does
    not follow the expected route format).
    """


class DuplicatedRouteError(Exception):
    """
    Exception for when an endpoint route is already in the added endpoints.
    """
